# -*- coding: utf-8 -*-

from enum import Enum

from qqbot.core.network.ws.ws_event import WsEvent
from qqbot.core.network.ws.ws_handler import DefaultHandler


class Handler:
    def __init__(self, handler_type, callback):
        self.type = handler_type
        self.callback = callback


def register_handlers(handlers):
    """
    RegisterHandlers 注册事件回调，并返回 intent 用于 websocket 的鉴权
    """
    intent = 0

    for handler in handlers:
        call_handler = intent_handler_dict.get(handler.type.value)
        intent = intent | call_handler(handler.callback, intent)

    return intent


def plain_event_handler(callback, intent):
    DefaultHandler.plain = callback
    return intent


def guild_event_handler(callback, intent):
    DefaultHandler.guild = callback
    intent = intent | WsEvent.event_to_intent(
        WsEvent.EventGuildCreate, WsEvent.EventGuildDelete, WsEvent.EventGuildUpdate
    )
    return intent


def guild_member_event_handler(callback, intent):
    DefaultHandler.guild_member = callback
    intent = intent | WsEvent.event_to_intent(
        WsEvent.EventGuildMemberAdd,
        WsEvent.EventGuildMemberRemove,
        WsEvent.EventGuildMemberUpdate,
    )
    return intent


def audio_event_handler(callback, intent):
    DefaultHandler.audio = callback
    intent = intent | WsEvent.event_to_intent(
        WsEvent.EventAudioStart,
        WsEvent.EventAudioFinish,
        WsEvent.EventAudioOnMic,
        WsEvent.EventAudioOffMic,
    )
    return intent


def channel_event_handler(callback, intent):
    DefaultHandler.channel = callback
    intent = intent | WsEvent.event_to_intent(
        WsEvent.EventChannelCreate,
        WsEvent.EventChannelDelete,
        WsEvent.EventChannelUpdate,
    )
    return intent


def message_event_handler(callback, intent):
    DefaultHandler.message = callback
    intent = intent | WsEvent.event_to_intent(WsEvent.EventMessageCreate)
    return intent


def at_message_event_handler(callback, intent):
    DefaultHandler.at_message = callback
    intent = intent | WsEvent.event_to_intent(WsEvent.EventAtMessageCreate)
    return intent


def direct_message_event_handler(callback, intent):
    DefaultHandler.direct_message = callback
    intent = intent | WsEvent.event_to_intent(WsEvent.EventDirectMessageCreate)
    return intent


def message_reactions_event_handler(callback, intent):
    DefaultHandler.message_reaction = callback
    intent = intent | WsEvent.event_to_intent(
        WsEvent.EventMessageReactionAdd,
        WsEvent.EventMessageReactionRemove,
    )
    return intent


class HandlerType(Enum):
    PLAIN_EVENT_HANDLER = 0
    GUILD_EVENT_HANDLER = 1
    GUILD_MEMBER_EVENT_HANDLER = 2
    CHANNEL_EVENT_HANDLER = 3
    MESSAGE_EVENT_HANDLER = 4
    AT_MESSAGE_EVENT_HANDLER = 5
    DIRECT_MESSAGE_EVENT_HANDLER = 6
    AUDIO_EVENT_HANDLER = 7
    MESSAGE_REACTIONS_EVENT_HANDLER = 8


intent_handler_dict = {
    HandlerType.PLAIN_EVENT_HANDLER.value: plain_event_handler,
    HandlerType.GUILD_EVENT_HANDLER.value: guild_event_handler,
    HandlerType.GUILD_MEMBER_EVENT_HANDLER.value: guild_member_event_handler,
    HandlerType.CHANNEL_EVENT_HANDLER.value: channel_event_handler,
    HandlerType.MESSAGE_EVENT_HANDLER.value: message_event_handler,
    HandlerType.AT_MESSAGE_EVENT_HANDLER.value: at_message_event_handler,
    HandlerType.DIRECT_MESSAGE_EVENT_HANDLER.value: direct_message_event_handler,
    HandlerType.AUDIO_EVENT_HANDLER.value: audio_event_handler,
    HandlerType.MESSAGE_REACTIONS_EVENT_HANDLER.value: message_reactions_event_handler,
}



# -*- coding: utf-8 -*-
import asyncio
import json

from qqbot.core.network.ws.ws_event import WsEvent
from qqbot.core.network.ws.ws_handler import DefaultHandler
from qqbot.core.util import logging
from qqbot.model.audio import AudioAction
from qqbot.model.channel import Channel
from qqbot.model.guild import Guild
from qqbot.model.guild_member import GuildMember
from qqbot.model.message import Message
from qqbot.model.reaction import Reaction

logger = logging.getLogger()


async def parse_and_handle(message_event, message):
    event_type = message_event["t"]
    call_handler = event_handler_dict.get(event_type)

    if call_handler is None:
        if DefaultHandler.plain is not None:
            plain = DefaultHandler.plain
            asyncio.ensure_future(plain(event_type, message))
        else:
            return
    else:
        asyncio.ensure_future(call_handler(event_type, message))


async def _handle_event_guild(message_event, message):
    callback = DefaultHandler.guild
    if callback is None:
        return
    guild: Guild = json.loads(_parse_data(message), object_hook=Guild)
    await callback(message_event, guild)


async def _handle_event_channel(message_event, message):
    callback = DefaultHandler.channel
    if callback is None:
        return
    channel: Channel = json.loads(_parse_data(message), object_hook=Channel)
    await callback(message_event, channel)


async def _handle_event_guild_member(message_event, message):
    callback = DefaultHandler.guild_member
    if callback is None:
        return
    guild_member: GuildMember = json.loads(
        _parse_data(message), object_hook=GuildMember
    )
    await callback(message_event, guild_member)


async def _handle_event_at_message_create(message_event, message):
    callback = DefaultHandler.at_message
    if callback is None:
        return
    at_message: Message = json.loads(_parse_data(message), object_hook=Message)
    await callback(message_event, at_message)


async def _handle_event_message_create(message_event, message):
    callback = DefaultHandler.message
    if callback is None:
        return
    msg: Message = json.loads(_parse_data(message), object_hook=Message)
    await callback(message_event, msg)


async def _handle_event_direct_message_create(message_event, message):
    callback = DefaultHandler.direct_message
    if callback is None:
        return
    msg: Message = json.loads(_parse_data(message), object_hook=Message)
    await callback(message_event, msg)


async def _handle_event_audio(message_event, message):
    callback = DefaultHandler.audio
    if callback is None:
        return
    audio_action: AudioAction = json.loads(
        _parse_data(message), object_hook=AudioAction
    )
    await callback(message_event, audio_action)


async def _handle_event_message_reaction(message_event, message):
    callback = DefaultHandler.message_reaction
    if callback is None:
        return
    reaction: Reaction = json.loads(_parse_data(message), object_hook=Reaction)
    await callback(message_event, reaction)


event_handler_dict = {
    WsEvent.EventGuildCreate: _handle_event_guild,
    WsEvent.EventGuildUpdate: _handle_event_guild,
    WsEvent.EventGuildDelete: _handle_event_guild,
    WsEvent.EventChannelCreate: _handle_event_channel,
    WsEvent.EventChannelUpdate: _handle_event_channel,
    WsEvent.EventChannelDelete: _handle_event_channel,
    WsEvent.EventGuildMemberAdd: _handle_event_guild_member,
    WsEvent.EventGuildMemberUpdate: _handle_event_guild_member,
    WsEvent.EventGuildMemberRemove: _handle_event_guild_member,
    WsEvent.EventAtMessageCreate: _handle_event_at_message_create,
    WsEvent.EventMessageCreate: _handle_event_message_create,
    WsEvent.EventDirectMessageCreate: _handle_event_direct_message_create,
    WsEvent.EventAudioStart: _handle_event_audio,
    WsEvent.EventAudioFinish: _handle_event_audio,
    WsEvent.EventAudioOnMic: _handle_event_audio,
    WsEvent.EventAudioOffMic: _handle_event_audio,
    WsEvent.EventMessageReactionAdd: _handle_event_message_reaction,
    WsEvent.EventMessageReactionRemove: _handle_event_message_reaction,
}


def _parse_data(message):
    json_obj = json.loads(message)
    if "d" in json_obj.keys():
        data = json_obj["d"]
        return json.dumps(data)



# -*- coding: utf-8 -*-

from qqbot.core.network.ws.dto.enum_intents import Intents
from qqbot.model.token import Token


class WSPayload:
    def __init__(self, d, op):
        self.op = op
        self.d = d


class WsIdentifyData:
    def __init__(self, token=Token, intents=Intents, shard=None):
        if shard is None:
            shard = []
        self.shard = shard
        self.token = token
        self.intents = intents


class WSResumeData:
    def __init__(self, token, session_id, seq):
        self.session_id = session_id
        self.token = token
        self.seq = seq


class Message:
    pass


class WSMessageData(Message):
    pass



# -*- coding: utf-8 -*-

from qqbot.model.token import Token


class Intent:
    pass


class ShardConfig:
    def __init__(self, shard_id, shard_count):
        self.shard_id = shard_id
        self.shard_count = shard_count


class Session:
    def __init__(
        self, session_id, url, intent, last_seq, token=Token, shards=ShardConfig
    ):
        self.shards = shards
        self.last_seq = last_seq
        self.intent = intent
        self.token = token
        self.url = url
        self.session_id = session_id



# -*- coding: utf-8 -*-
from typing import List


class APIPermission:
    def __init__(self, data=None):
        self.path: str = ""
        self.method: str = ""
        self.desc: str = ""
        self.auth_status: int
        if data:
            self.__dict__ = data


class APIPermissionDemandIdentify:
    def __init__(self, path: str, method: str):
        self.path = path
        self.method = method


class APIPermissionDemand:
    def __init__(
        self,
        data=None,
        guild_id: str = None,
        channel_id: str = None,
        api_identify: APIPermissionDemandIdentify = None,
        title: str = None,
        desc: str = None,
    ):
        if guild_id:
            self.guild_id = guild_id
        if channel_id:
            self.channel_id = channel_id
        if api_identify:
            self.api_identify = api_identify
        if title:
            self.title = title
        if desc:
            self.desc = desc
        if data:
            self.__dict__ = data


class PermissionDemandToCreate:
    def __init__(
        self, channel_id: str, api_identify: APIPermissionDemandIdentify, desc: str = ""
    ):
        self.channel_id = channel_id
        self.api_identify = api_identify
        self.desc = desc


class APIs:
    def __init__(self, data=None):
        self.apis: List[APIPermission]
        if data:
            self.__dict__ = data



# -*- coding: utf-8 -*-


class ReqOption:
    def __init__(self, before: str = None, after: str = None, limit: int = None):
        """
        获取频道列表的请求

        :param before: 读取此 guild id 之前的数据
        :param after: 读取此 guild id 之后的数据
        :param limit: 每次拉取多少条数据，最大不超过 100，默认 100
        """
        if before is not None:
            self.before = before
        if after is not None:
            self.after = after
        if limit is not None:
            self.limit = limit



# -*- coding: utf-8 -*-

from qqbot.core.network.ws.dto.enum_intents import Intents


class WsEvent:
    EventGuildCreate = "GUILD_CREATE"
    EventGuildUpdate = "GUILD_UPDATE"
    EventGuildDelete = "GUILD_DELETE"
    EventChannelCreate = "CHANNEL_CREATE"
    EventChannelUpdate = "CHANNEL_UPDATE"
    EventChannelDelete = "CHANNEL_DELETE"

    EventGuildMemberAdd = "GUILD_MEMBER_ADD"
    EventGuildMemberUpdate = "GUILD_MEMBER_UPDATE"
    EventGuildMemberRemove = "GUILD_MEMBER_REMOVE"

    EventMessageCreate = "MESSAGE_CREATE"
    EventAtMessageCreate = "AT_MESSAGE_CREATE"
    EventDirectMessageCreate = "DIRECT_MESSAGE_CREATE"

    EventAudioStart = "AUDIO_START"
    EventAudioFinish = "AUDIO_FINISH"
    EventAudioOnMic = "AUDIO_ON_MIC"
    EventAudioOffMic = "AUDIO_OFF_MIC"

    EventMessageReactionAdd = "MESSAGE_REACTION_ADD"
    EventMessageReactionRemove = "MESSAGE_REACTION_REMOVE"

    # 事件map
    event_dirt = {
        EventGuildCreate: Intents.INTENT_GUILDS,
        EventGuildUpdate: Intents.INTENT_GUILDS,
        EventGuildDelete: Intents.INTENT_GUILDS,
        EventChannelCreate: Intents.INTENT_GUILDS,
        EventChannelUpdate: Intents.INTENT_GUILDS,
        EventChannelDelete: Intents.INTENT_GUILDS,
        EventGuildMemberAdd: Intents.INTENT_GUILD_MEMBERS,
        EventGuildMemberUpdate: Intents.INTENT_GUILD_MEMBERS,
        EventGuildMemberRemove: Intents.INTENT_GUILD_MEMBERS,
        EventMessageCreate: Intents.INTENT_GUILD_MESSAGES,
        EventAtMessageCreate: Intents.INTENT_GUILD_AT_MESSAGE,
        EventDirectMessageCreate: Intents.INTENT_DIRECT_MESSAGE,
        EventAudioStart: Intents.INTENT_AUDIO,
        EventAudioFinish: Intents.INTENT_AUDIO,
        EventAudioOnMic: Intents.INTENT_AUDIO,
        EventAudioOffMic: Intents.INTENT_AUDIO,
        EventMessageReactionAdd: Intents.INTENT_MESSAGE_REACTION,
        EventMessageReactionRemove: Intents.INTENT_MESSAGE_REACTION,
    }

    @staticmethod
    def event_to_intent(*events):
        intent = 0
        for event in events:
            intent = intent | WsEvent.event_dirt[event].value
        return intent

    @staticmethod
    def intent_to_event(intent_event_map):
        pass



from qqbot.model.emoji import Emoji


class Reaction:
    def __init__(self, data=None):
        self.user_id: str = ""
        self.guild_id: str = ""
        self.channel_id: str = ""
        self.target: ReactionTarget = ReactionTarget()
        self.emoji: Emoji = Emoji()
        if data:
            self.__dict__ = data


class ReactionTarget:
    def __init__(self, data=None):
        self.id: str = ""
        self.type: int = 0
        if data:
            self.__dict__ = data


class ReactionTargetType:
    message = 0
    feed = 1
    comment = 2
    reply = 3



# -*- coding: utf-8 -*-
import json
import threading
import traceback

import websocket

from qqbot.core.exception.error import WebsocketError
from qqbot.core.network.ws.dto.enum_intents import Intents
from qqbot.core.network.ws.dto.enum_opcode import OpCode
from qqbot.core.network.ws.dto.ws_payload import (
    WSPayload,
    WsIdentifyData,
    WSResumeData,
)
from qqbot.core.network.ws_sync.ws_event_handler import parse_and_handle
from qqbot.core.util import logging

logger = logging.getLogger()


def _loop_exception_handler(loop, context):
    # first, handle with default handler
    loop.default_exception_handler(context)

    exception = context.get("exception")
    if isinstance(exception, ZeroDivisionError):
        print(context)
        loop.stop()


class Client:
    def __init__(self, session, session_manager):
        self.session = session
        self.ws_conn = None
        self.heartbeat_thread = None
        self.session_manager = session_manager
        self.can_reconnect = False

    def connect(self, connected_callback):
        """
        websocket向服务器端发起链接，并定时发送心跳
        :param connected_callback:链接成功后的回调
        """
        ws_url = self.session.url
        if ws_url == "":
            raise Exception("session url is none")

        def on_close(ws, close_status_code, close_msg):
            logger.info(
                "on_close: websocket connection %s" % ws
                + ", code: %s" % close_status_code
                + ", msg:%s" % close_msg
            )
            # 关闭心跳包线程
            self.ws_conn = None
            # 这种不能重新链接
            if (
                close_status_code == WebsocketError.CodeConnCloseErr
                or close_status_code == WebsocketError.CodeInvalidSession
                or self.can_reconnect is False
            ):
                self.session.session_id = ""
                self.session.last_seq = 0
            # 断连后启动一个新的链接并透传当前的session，不使用内部重连的方式，避免死循环
            self.session_manager.session_pool.add_task(self.session)
            self.session_manager.start_session()

        def on_message(ws, message):
            logger.debug("on_message: %s" % message)
            message_event = json.loads(message)
            if self._is_system_event(message_event, ws, connected_callback):
                return
            if "t" in message_event.keys() and message_event["t"] == "READY":
                event_seq = message_event["s"]
                if event_seq > 0:
                    self.session.last_seq = event_seq
                self._ready_handler(message_event)
                return
            if "t" in message_event.keys():
                parse_and_handle(message_event, message)

        def on_error(ws, exception=Exception):
            traceback.print_exc()

        def on_open(ws):
            logger.info("on_open: %s" % ws)

        ws_app = websocket.WebSocketApp(
            ws_url,
            on_message=on_message,
            on_error=on_error,
            on_close=on_close,
            on_open=on_open,
        )

        self.heartbeat_thread = threading.Thread(
            target=self._send_heartbeat, args=(30, (threading.Event()))
        )
        self.heartbeat_thread.daemon = True
        self.heartbeat_thread.start()

        ws_app.run_forever()

    def identify(self):
        """
        websocket鉴权
        """
        if self.session.intent == 0:
            self.session.intent = Intents.INTENT_GUILDS.value
        logger.info("ws:%s is identify" % self.ws_conn)
        identify_event = json.dumps(
            WSPayload(
                WsIdentifyData(
                    token=self.session.token.get_string(),
                    intents=self.session.intent,
                    shard=[
                        self.session.shards.shard_id,
                        self.session.shards.shard_count,
                    ],
                ).__dict__,
                op=OpCode.WS_IDENTITY.value,
            ).__dict__
        )
        self.send_msg(identify_event)

    def send_msg(self, event_json):
        """
        websocket发送消息
        :param event_json:
        """
        send_msg = event_json
        logger.debug("send_msg: %s" % send_msg)
        self.ws_conn.send(data=send_msg)

    def reconnect(self):
        """
        websocket重连
        """
        logger.info("ws:%s is reconnected" % self.ws_conn)
        resume_event = json.dumps(
            WSPayload(
                WSResumeData(
                    token=self.session.token.get_string(),
                    session_id=self.session.session_id,
                    seq=self.session.last_seq,
                ).__dict__,
                op=OpCode.WS_RESUME.value,
            ).__dict__
        )
        self.send_msg(resume_event)

    def _ready_handler(self, message_event):
        data = message_event["d"]
        self.version = data["version"]
        self.session.session_id = data["session_id"]
        self.session.shards.shard_id = data["shard"][0]
        self.session.shards.shard_count = data["shard"][1]
        self.user = data["user"]

    def _is_system_event(self, message_event, ws, connected_callback):
        """
        系统事件
        :param message_event:消息
        :param ws:websocket
        :param connected_callback:回调
        :return:
        """
        event_op = message_event["op"]
        if event_op == OpCode.WS_HELLO.value:
            self.ws_conn = ws
            connected_callback(self)
            return True
        if event_op == OpCode.WS_HEARTBEAT.value:
            return True
        if event_op == OpCode.WS_RECONNECT.value:
            self.can_reconnect = True
            return True
        if event_op == OpCode.WS_INVALID_SESSION.value:
            self.can_reconnect = False
            return True
        return False

    def _send_heartbeat(self, interval, thread_event):
        """
        心跳包
        :param interval: 间隔时间
        :param thread_event: 线程
        """
        while not thread_event.wait(interval):
            heartbeat_event = json.dumps(
                WSPayload(
                    op=OpCode.WS_HEARTBEAT.value, d=self.session.last_seq
                ).__dict__
            )
            if self.ws_conn is None:
                self.heartbeat_thread.stopped = True
            else:
                self.send_msg(heartbeat_event)



# -*- coding: utf-8 -*-

import logging
import os
import platform
from logging import FileHandler
from logging.handlers import TimedRotatingFileHandler

LOG_COLORS_CONFIG = {
    "DEBUG": "cyan",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "red",
}


def _getLevel():
    level = logging.INFO
    level_str = os.getenv("QQBOT_LOG_LEVEL", str(logging.INFO))
    try:
        level = int(level_str)
        if level not in (
            logging.NOTSET,
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
            logging.CRITICAL,
        ):
            logging.error("wrong logging level %s" % level_str)
            level = logging.INFO
    except ValueError:
        logging.error("wrong logging level %s" % level_str)
    logging.info("logging level: %d" % level)
    return level


def getLogger(name=None):
    print_format = (
        "%(asctime)s - "
        "\033[1;33m%(levelname)s: %(name)s - %(filename)s - %(funcName)s(line: %(lineno)s):\033[0m%(message)s"
        ""
    )
    file_format = "%(asctime)s-%(name)s - %(filename)s - %(funcName)s - line %(lineno)s-%(levelname)s - %(message)s"

    if name is None:
        logger = logging.getLogger("qqbot")
    else:
        logger = logging.getLogger(name)

    logging.basicConfig(format=print_format)
    logger.setLevel(level=_getLevel())

    # FileHandler
    no_log = os.getenv("QQBOT_DISABLE_LOG", "0")
    if no_log == "0":
        formatter = logging.Formatter(file_format)
        if name is None:
            name = "qqbot"
        log_file = os.path.join(os.getcwd(), name + ".log")
        file_handler = None
        if platform.system().lower() != "windows":
            # do not use RotatingFileHandler under Windows
            # due to multi-process issue
            # file_handler = RotatingFileHandler(
            #     log_file,
            #     maxBytes=1024 * 1024,
            #     backupCount=5,
            # )
            # save last 7 days log
            file_handler = TimedRotatingFileHandler(
                filename=log_file,
                when="D",
                backupCount=7,
            )
        else:
            file_handler = FileHandler(log_file, encoding="utf-8")
        logger.debug(
            "qqbot: dumping log file to {path}".format(path=os.path.realpath(log_file))
        )
        file_handler.setLevel(level=_getLevel())
        file_handler.setFormatter(formatter)
        if len(logger.handlers) == 0:
            logger.addHandler(file_handler)

    return logger
